--[[--------------------------------------------------------------------
	OnScreenHealth
	Minimal text-based HUD for the player, pet, and target units.
	by Phanx < addons@phanx.net >
	Copyright  20072010 Phanx. Some rights reserved. See LICENSE.txt for details.
	http://www.wowinterface.com/downloads/info7470-OnScreenHealth.html
	http://wow.curse.com/downloads/wow-addons/details/onscreenhealth.aspx
----------------------------------------------------------------------]]

-- --------------------------------------------------------------------
-- ADDON OBJECTS
-- --------------------------------------------------------------------
local ADDON_NAME, ns = ...
local VERSION = GetAddOnMetadata(ADDON_NAME,"Version")

local OnScreenHealth = CreateFrame("Frame", "OnScreenHealth")
OnScreenHealth:SetScript("OnEvent", function(f, e, ...) return f[e] and f[e](f, ...) end)
OnScreenHealth:RegisterEvent("ADDON_LOADED")

ns.OnScreenHealth = OnScreenHealth

local db, currentToT

local unitFrames = { }
ns.unitFrames = unitFrames

local SharedMedia = LibStub("LibSharedMedia-3.0")

-- --------------------------------------------------------------------
-- GLOBALS UPVALUED
-- --------------------------------------------------------------------
local UnitExists, UnitName, UnitIsConnected, UnitIsGhost, UnitIsDead, UnitIsDeadOrGhost, UnitIsPlayer =
	UnitExists, UnitName, UnitIsConnected, UnitIsGhost, UnitIsDead, UnitIsDeadOrGhost, UnitIsPlayer
local UnitHealth, UnitHealthMax, UnitPower, UnitPowerMax, UnitPowerType, GetSpecialization, GetEclipseDirection =
	UnitHealth, UnitHealthMax, UnitPower, UnitPowerMax, UnitPowerType, GetSpecialization, GetEclipseDirection
local pairs, unpack, format, gsub, abs = pairs, unpack, format, gsub, abs

-- --------------------------------------------------------------------
-- LOCALIZATION
-- --------------------------------------------------------------------
if not ns.L then
	ns.L = { }
end

local L = setmetatable(ns.L, {
	__index = function(t, k)
		t[k] = k
		return k
	end
})

L["Health"] = HEALTH
L["Mana"] = MANA
L["Rage"] = RAGE
L["Focus"] = FOCUS
L["Energy"] = ENERGY
L["Runic Power"] = RUNIC_POWER
L["Combo Points"] = COMBAT_TEXT_SHOW_COMBO_POINTS_TEXT
L["Runes"] = RUNES
L["Blood Rune"] = COMBAT_TEXT_RUNE_BLOOD
L["Unholy Rune"] = COMBAT_TEXT_RUNE_UNHOLY
L["Frost Rune"] = COMBAT_TEXT_RUNE_FROST
L["Death Rune"] = COMBAT_TEXT_RUNE_DEATH
L["Soul Shards"] = SOUL_SHARDS_POWER
L["Eclipse"] = ECLIPSE
L["Solar Energy"] = BALANCE_POSITIVE_ENERGY
L["Lunar Energy"] = BALANCE_NEGATIVE_ENERGY
L["Holy Power"] = HOLY_POWER
L["Chi"] = CHI
L["Shadow Orbs"] = SHADOW_ORBS
L["Burning Embers"] = BURNING_EMBERS
L["Demonic Fury"] = DEMONIC_FURY

L["<<YOU>>"] = "<<"..YOU..">>"
L["Pet"] = PET
L["Player"] = PLAYER
L["Target"] = TARGET
L["Dead"] = DEAD
L["Offline"] = PLAYER_OFFLINE
L["Enable"] = ENABLE
L["Mode"] = MODE
L["None"] = NONE
L["Appearance"] = APPEARANCE_LABEL
L["Colors"] = COLORS

-- --------------------------------------------------------------------
-- STATIC DATA
-- --------------------------------------------------------------------
local playerUnit = "player"
local playerClass = select(2, UnitClass("player"))

local RAID_CLASS_COLORS = CUSTOM_CLASS_COLORS or _G.RAID_CLASS_COLORS

local realPowerTypes = {
	["MANA"] = true,
	["RAGE"] = true,
	["FOCUS"] = true,
	["ENERGY"] = true,
	["RUNIC_POWER"] = true,
	["AMMOSLOT"] = true,
	["FUEL"] = true,
}
-- class, spec, 0 = all specs
local resourceTypes = {
	["COMBO_POINTS"] = {["DRUID"] = 0, ["ROGUE"] = 0},
	["RUNES"] = {["DEATHKNIGHT"] = 0},
	["SOUL_SHARDS"] = {["WARLOCK"] = SPEC_WARLOCK_AFFLICTION},
	["ECLIPSE"] = {["DRUID"] = 1},
	["HOLY_POWER"] = {["PALADIN"] = 0},
	["DARK_FORCE"] = {["MONK"] = 0},
	["CHI"] = {["MONK"] = 0},
	["SHADOW_ORBS"] = {["PRIEST"] = SPEC_PRIEST_SHADOW},
	["BURNING_EMBERS"] = {["WARLOCK"] = SPEC_WARLOCK_DESTRUCTION},
	["DEMONIC_FURY"] = {["WARLOCK"] = SPEC_WARLOCK_DEMONOLOGY},
}
local powerTypeID = {
	["MANA"] = SPELL_POWER_MANA,
	["RAGE"] = SPELL_POWER_RAGE,
	["FOCUS"] = SPELL_POWER_FOCUS,
	["ENERGY"] = SPELL_POWER_ENERGY,
	["RUNES"] = SPELL_POWER_RUNES,
	["RUNIC_POWER"] = SPELL_POWER_RUNIC_POWER,
	["SOUL_SHARDS"] = SPELL_POWER_SOUL_SHARDS,
	["ECLIPSE"] = SPELL_POWER_ECLIPSE,
	["HOLY_POWER"] = SPELL_POWER_HOLY_POWER,
	["DARK_FORCE"] = SPELL_POWER_DARK_FORCE,
	["CHI"] = SPELL_POWER_CHI,
	["SHADOW_ORBS"] = SPELL_POWER_SHADOW_ORBS,
	["BURNING_EMBERS"] = SPELL_POWER_BURNING_EMBERS,
	["DEMONIC_FURY"] = SPELL_POWER_DEMONIC_FURY,
	["AMMOSLOT"] = SPELL_POWER_AMMOSLOT,
	["FUEL"] = SPELL_POWER_FUEL,
}
local RUNETYPE_BLOOD,RUNETYPE_UNHOLY,RUNETYPE_FROST,RUNETYPE_DEATH, MAX_RUNES = 1,2,3,4,6
local runeColors = {
	[RUNETYPE_BLOOD] = {1, 0, 0},
	[RUNETYPE_UNHOLY] = {0, 0.5, 0},
	[RUNETYPE_FROST] = {0, 1, 1},
	[RUNETYPE_DEATH] = {0.8, 0.1, 1},
}

-- --------------------------------------------------------------------
-- DYNAMIC DATA
-- --------------------------------------------------------------------
local units = {
	["pet"]     = true,
	["player"]  = true,
	["target"]  = true,
	["vehicle"] = true,
}
ns.units = units
local spec, hasLevelReq, hasSpellReq = nil, true, true

-- --------------------------------------------------------------------
-- UTILITY FUNCTIONS
-- --------------------------------------------------------------------
local function Debug(lvl, str, ...)
	if lvl > 0 then return end
	if (...) then
		if str:match("%%") then
			str = str:format(...)
		else
			str = string.join(", ", str, tostringall(...))
		end
	end
	return print("|cffff6666[DEBUG] OnScreenHealth:|r", str)
end
ns.Debug = Debug

local function si(v)
	if not v then return "" end
	if type(v) ~= "number" then return v end
	Debug(5, "si(%d)", v)

	if v >= 10000000 then
		return ("%.1fm"):format(v / 1000000)
	elseif v >= 1000000 then
		return ("%.2fm"):format(v / 1000000)
	elseif v >= 100000 then
		return ("%.0fk"):format(v / 1000)
	elseif v >= 10000 then
		return ("%.1fk"):format(v / 1000)
	else
		return v
	end
end

local function GetUnitColor(unit)
	if not unit then return end

	if UnitIsDeadOrGhost(unit) or not UnitIsConnected(unit) then
		return db.colorAbsent
	else
		local color = UnitIsPlayer(unit) and "|c"..RAID_CLASS_COLORS[select(2, UnitClass(unit))].colorStr or nil
		if color then
			return color
		else
			local r, g, b = GameTooltip_UnitColor(unit)
			return string.format("|cff%02x%02x%02x", r * 255, g * 255, b * 255)
		end
	end
end

local function GetFormattedText(text, cur, max)
	if not text and cur and max then return end
	Debug(3, "GetFormattedText")

	text = text:gsub("$c:s", si(cur))
	text = text:gsub("$c", cur)

	text = text:gsub("$m:s", si(max))
	text = text:gsub("$m", max)

	text = text:gsub("$d:s", si(max - cur))
	text = text:gsub("$d", max - cur)

	text = text:gsub("$p", floor(cur / max * 100))

	return text
end

-- --------------------------------------------------------------------
-- EVENT PROCESSING
-- --------------------------------------------------------------------
local runeAmounts = {}
function OnScreenHealth:RUNE_POWER_UPDATE()
	if not db.showResource then return end
	wipe(runeAmounts)
	for i=1,MAX_RUNES do
		local runeType = GetRuneType(i)
		local _, _, runeReady = GetRuneCooldown(i)
		if runeType and runeReady then
			runeAmounts[runeType] = runeAmounts[runeType] and (runeAmounts[runeType] + 1) or 1
		end
	end
	if next(runeAmounts) then
		local colors = db.resourceColors
		unitFrames.player.Resource:SetFormattedText("%s%d|r:%s%d|r:%s%d|r:%s%d|r",colors["RUNES"..RUNETYPE_BLOOD],runeAmounts[RUNETYPE_BLOOD] or 0,colors["RUNES"..RUNETYPE_UNHOLY],runeAmounts[RUNETYPE_UNHOLY] or 0,colors["RUNES"..RUNETYPE_FROST],runeAmounts[RUNETYPE_FROST] or 0,colors["RUNES"..RUNETYPE_DEATH],runeAmounts[RUNETYPE_DEATH] or 0)
	else
 		unitFrames.player.Resource:SetText()
	end
end
OnScreenHealth.RUNE_TYPE_UPDATE = OnScreenHealth.RUNE_POWER_UPDATE

function OnScreenHealth:UNIT_PET(owner)
	if owner and owner ~= "player" or db.disablePet then return end
	Debug(1, "UNIT_PET")

	if playerUnit == "vehicle" or not db.petEnable or not (db.playerShowOOC or UnitAffectingCombat(playerUnit)) or not (db.petShowDead or not UnitIsDeadOrGhost("pet")) or not UnitExists("pet") then
		unitFrames.pet:Hide()
		return
	end

	if UnitExists("pet") then
		if db.hidePetDead then
			if UnitIsDeadOrGhost("pet") then
				unitFrames.pet:Hide()
				return
			end
		end
		unitFrames.pet:Show()
	else
		unitFrames.pet:Hide()
	end
end

function OnScreenHealth:PLAYER_TARGET_CHANGED()
	Debug(1, "PLAYER_TARGET_CHANGED")
	if not UnitExists("target") or not db.targetEnable or UnitIsUnit("target",playerUnit) or not (db.targetShowOOC or UnitAffectingCombat(playerUnit)) or not (db.targetShowDead or (not UnitIsDeadOrGhost("target") and UnitIsConnected("target"))) then
		unitFrames.target:Hide()
		currentToT = nil
		return
	end

	unitFrames.target:Show()

	if playerClass == "ROGUE" or playerClass == "DRUID" then
		self:UNIT_COMBO_POINTS()
	end
end

function OnScreenHealth:UNIT_COMBO_POINTS()
	if not db.showResource then return end
	if playerClass == "DRUID" then
		local form  = GetShapeshiftFormID()
		if not form or form ~= CAT_FORM then
			unitFrames.target.ComboPoints:SetText()
			return
		end
	end
	local cp = GetComboPoints(playerUnit, "target")
	if cp > 0 then
		unitFrames.target.ComboPoints:SetFormattedText("%s%s|r", db.resourceColors["COMBO_POINTS"], cp)
	else
		unitFrames.target.ComboPoints:SetText()
	end
end

function OnScreenHealth:UNIT_ENTERED_VEHICLE(unit)
	Debug(1, "UNIT_ENTERED_VEHICLE", unit)
	if unit ~= "player" then return end
	playerUnit = "vehicle"
	self:UNIT_PET("player")
	ns.units["player"] = false
	unitFrames.player.Resource:SetText()
end

function OnScreenHealth:UNIT_EXITED_VEHICLE(unit)
	Debug(1, "UNIT_EXITED_VEHICLE", unit)
	if unit ~= "player" then return end
	playerUnit = "player"
	self:UNIT_PET("player")
	ns.units["player"] = true
end

function OnScreenHealth:PLAYER_REGEN_DISABLED()
	Debug(1, "PLAYER_REGEN_DISABLED")
	self:UpdatePlayer()
	self:UNIT_PET("player")
	self:PLAYER_TARGET_CHANGED()
	if db.alphaMode == "COMBAT" then
		local a = db.alphaCombat
		unitFrames.player:SetAlpha(a)
		unitFrames.target:SetAlpha(a)
		unitFrames.pet:SetAlpha(a)
	end
end

function OnScreenHealth:PLAYER_REGEN_ENABLED()
	Debug(1, "PLAYER_REGEN_ENABLED")
	self:UpdatePlayer()
	if not db.playerShowOOC then
		unitFrames.player:Hide()
		unitFrames.pet:Hide()
	end
	if not db.targetShowOOC then
		self:PLAYER_TARGET_CHANGED()
	end
	if db.alphaMode == "COMBAT" then
		local a = db.alphaOOC
		unitFrames.player:SetAlpha(a)
		unitFrames.target:SetAlpha(a)
		unitFrames.pet:SetAlpha(a)
	end
end

function OnScreenHealth:PLAYER_DEAD()
	Debug(1, "PLAYER_DEAD")
	self:UpdateCombatStatus(playerUnit)
	unitFrames.player:Hide()
	unitFrames.pet:Hide()
end

function OnScreenHealth:PLAYER_ALIVE()
	Debug(1, "PLAYER_ALIVE")
	if not UnitIsGhost(playerUnit) and db.playerShowOOC or UnitAffectingCombat(playerUnit) then
		unitFrames.player:Show()
	end
end
OnScreenHealth.PLAYER_UNGHOST = OnScreenHealth.PLAYER_ALIVE -- redundant since it registers _UNGHOST to _ALIVE handler

function OnScreenHealth:PLAYER_TALENT_UPDATE() -- warlock, priest, druid
	if playerClass == "WARLOCK" or playerClass == "PRIEST" or playerClass == "DRUID" then
		spec = GetSpecialization()
		if playerClass == "WARLOCK" then
			if not spec then
				hasSpellReq = false
			elseif spec == SPEC_WARLOCK_AFFLICTION then
				if IsPlayerSpell(WARLOCK_SOULBURN) then
					hasSpellReq = true
					self:UnregisterEvent("SPELLS_CHANGED")
				else
					hasSpellReq = false
					self:RegisterEvent("SPELLS_CHANGED")
				end
			elseif spec == SPEC_WARLOCK_DESTRUCTION then
				if IsPlayerSpell(WARLOCK_BURNING_EMBERS) then
					hasSpellReq = true
					self:UnregisterEvent("SPELLS_CHANGED")
				else
					hasSpellReq = false
					self:RegisterEvent("SPELLS_CHANGED")
				end
			elseif spec == SPEC_WARLOCK_DEMONOLOGY then
				hasSpellReq = true
				self:UnregisterEvent("SPELLS_CHANGED")
			end
		end
	end
end

function OnScreenHealth:PLAYER_LEVEL_UP() -- paladin, priest
	if playerClass == "PALADIN" or playerClass == "PRIEST" then
		local playerLevel = UnitLevel("player")
		if (playerClass == "PALADIN" and playerLevel >= PALADINPOWERBAR_SHOW_LEVEL) or (playerClass == "PRIEST" and playerLevel >= SHADOW_ORBS_SHOW_LEVEL) then
			hasLevelReq = true
			self:UnregisterEvent("PLAYER_LEVEL_UP")
		else
			hasLevelReq = false
		end
	end
end

function OnScreenHealth:SPELLS_CHANGED() -- warlock
	if not playerClass == "WARLOCK" then return end
	if not spec then
		hasSpellReq = false
	elseif spec == SPEC_WARLOCK_AFFLICTION then
		if IsPlayerSpell(WARLOCK_SOULBURN) then
			hasSpellReq = true
			self:UnregisterEvent("SPELLS_CHANGED")
		else
			hasSpellReq = false
		end
	elseif spec == SPEC_WARLOCK_DESTRUCTION then
		if IsPlayerSpell(WARLOCK_BURNING_EMBERS) then
			hasSpellReq = true
			self:UnregisterEvent("SPELLS_CHANGED")
		else
			hasSpellReq = false
		end
	elseif spec == SPEC_WARLOCK_DEMONOLOGY then
		hasSpellReq = true
		self:UnregisterEvent("SPELLS_CHANGED")
	end
end

function OnScreenHealth:UPDATE_SHAPESHIFT_FORM() -- druid
	if not db.showResource or not playerClass == "DRUID" then return end
	local form = GetShapeshiftFormID()
	if form == MOONKIN_FORM or not form then
		if spec and spec == 1 then -- balance
			self:UpdatePlayer()
		else
			unitFrames.player.Resource:SetText()
		end
		if not form then unitFrames.target.ComboPoints:SetText() end
	elseif form == CAT_FORM then
		self:UpdatePlayer()
	else
		unitFrames.player.Resource:SetText()
		unitFrames.target.ComboPoints:SetText()
	end
end

function OnScreenHealth:ADDON_LOADED(addon)
	if addon ~= ADDON_NAME then return end
	Debug(1, "ADDON_LOADED", addon)

	local defaults = {
		playerEnable = true,
		playerShowDead = false,
		playerShowOOC = true,
		playerShowCombatStatus = false,
		playerFrequentUpdates = 0.5,
		targetEnable = true,
		targetShowDead = true,
		targetShowOOC = true,
		targetShowCP = true,
		targetShowToT = false,
		petEnable = true,
		petShowDead = false,
		alphaMode = "FIXED",
		alphaFixed = 1,
		alphaCombat = 1,
		alphaOOC = 0.5,
		colorHealthMode = "GRADIENT",
		showPower = true,
		showResource = true,
		textAlign = "INSIDE",
		textFormatHealthMonster = "$c",
		textFormatHealthPlayer = "$c",
		textFormatPowerMonster = "$c",
		textFormatPowerPlayer = "$c",
		fontFace = "Friz Quadrata TT",
		fontOutline = "OUTLINE",
		fontShadow = true,
		fontSizeHealth = 32,
		fontSizePower = 24,
		fontSizeTarget = 24,
		fontSizeResource = 24,
		fontSizePet = 0.8,
		colorHealthMax = { 0, 1, 0 },
		colorHealthMid = { 1, 1, 0 },
		colorHealthMin = { 1, 0, 0 },
		powerColors = {},
		resourceColors = {
			["COMBO_POINTS"] = "|cffff7f00",
			["SHADOW_ORBS"] = "|cffcc99ff",
			["BURNING_EMBERS"] = "|cffCC3300",
			["DEMONIC_FURY"] = "|cff006600",
		},
		colorAbsent = "|cff999999",
		colorCombatStatus = "|cffffffff",
		colorTargetSelf = "|cffffffff",
		colorDefault = "|cffff00ff", -- Fallback
		posX = 200,
		posY = 0,
		posPet = 20,
		frameStrata = "BACKGROUND",
		version = VERSION
	}

	for k, v in pairs(PowerBarColor) do
		if type(k) == "string" and type(v) == "table" and v.r then
			if resourceTypes[k] then
				defaults.resourceColors[k] = ("|cff%02x%02x%02x"):format(v.r * 255, v.g * 255, v.b * 255)
			elseif realPowerTypes[k] then
				defaults.powerColors[k] = ("|cff%02x%02x%02x"):format(v.r * 255, v.g * 255, v.b * 255)
			end
		end
	end
	for k, v in pairs(PowerBarColor.ECLIPSE) do
		defaults.resourceColors[k == "positive" and "ECLIPSE_SOLAR" or "ECLIPSE_LUNAR"] = ("|cff%02x%02x%02x"):format(v.r * 255, v.g * 255, v.b * 255)
	end
	for k, v in pairs(runeColors) do
		defaults.resourceColors["RUNES"..k] = ("|cff%02x%02x%02x"):format(v[1] * 255, v[2] * 255, v[3] * 255)
	end

	self.db = LibStub("AceDB-3.0"):New("OnScreenHealthDB", { profile = defaults }, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "Refresh")
	self.db.RegisterCallback(self, "OnProfileCopied", "Refresh")
	self.db.RegisterCallback(self, "OnProfileReset", "Refresh")
	db = self.db.profile
 	if db.playerFrequentUpdates and type(db.playerFrequentUpdates)~="number" then
 		db.playerFrequentUpdates = defaults.playerFrequentUpdates
 	end -- ticket 8 (curseforge) pre 5.x SV

	self:LoadOptions(playerClass)

	self:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil

	if IsLoggedIn() then
		self:PLAYER_LOGIN()
	else
		self:RegisterEvent("PLAYER_LOGIN")
	end
end

function OnScreenHealth:PLAYER_LOGIN()
	Debug(1, "PLAYER_LOGIN")

	ns.units["player"] = db.playerEnable
	ns.units["vehicle"] = db.playerEnable
	ns.units["target"] = db.targetEnable
	ns.units["pet"] = db.petEnable

	if not unitFrames.player then
		self:CreateFrames()
	end

	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_ENTERED_VEHICLE")
	self:RegisterEvent("UNIT_EXITED_VEHICLE")
	if playerClass == "WARLOCK" or playerClass == "DRUID" or playerClass == "PRIEST" then
		self:RegisterEvent("PLAYER_TALENT_UPDATE")
		self:PLAYER_TALENT_UPDATE()
	end
	if playerClass == "PALADIN" or playerClass == "PRIEST" then
		self:RegisterEvent("PLAYER_LEVEL_UP")
		self:PLAYER_LEVEL_UP()
	end
	if playerClass == "WARLOCK" then
		self:RegisterEvent("SPELLS_CHANGED")
		self:SPELLS_CHANGED()
	end

	self:UpdateResourceEvents()
	self:UpdateCombatEvents()
	self:UpdateDeathEvents()

	self:RegisterEvent("UNIT_PET")

 	if db.playerEnable or db.showPower or db.showResource or db.targetShowToT then
 		self.updateframe.timer = 0
		self.updateframe:SetScript("OnUpdate", self.OnUpdate)
 	end

	self:UnregisterEvent("PLAYER_LOGIN")
end

-- --------------------------------------------------------------------
-- ADDON METHODS
-- --------------------------------------------------------------------
function OnScreenHealth.UpdateHealth()
	-- check options
	if not (db.playerEnable or db.targetEnable or db.petEnable) then return end
	for unit,_ in pairs(ns.units) do
		local f = unitFrames[unit]
		if ns.units[unit] and UnitExists(unit) and f and f:IsShown() then

			if not UnitIsConnected(unit) then
				f.Health:SetFormattedText("%s%s|r", db.colorAbsent, L["Offline"])
			elseif UnitIsGhost(unit) then
				f.Health:SetFormattedText("%s%s|r", db.colorAbsent, L["Ghost"])
			elseif UnitIsDead(unit) then
				f.Health:SetFormattedText("%s%s|r", db.colorAbsent, L["Dead"])
			end

			local cur, max = UnitHealth(unit), UnitHealthMax(unit)
			local percent = cur / max

			local color
			if db.colorHealthMode == "CLASS" then
				color = GetUnitColor(unit)
			elseif db.colorHealthMode == "GRADIENT" then
				local r1, g1, b1, r2, g2, b2
				if percent <= 0.5 then
					percent = percent * 2
					r1, g1, b1 = unpack(db.colorHealthMin)
					r2, g2, b2 = unpack(db.colorHealthMid)
				else
					percent = percent * 2 - 1
					r1, g1, b1 = unpack(db.colorHealthMid)
					r2, g2, b2 = unpack(db.colorHealthMax)
				end
				color = ("|cff%02x%02x%02x"):format((r1 + (r2 - r1) * percent) * 255, (g1 + (g2 - g1) * percent) * 255, (b1 + (b2 - b1) * percent) * 255)
			else
				local r, g, b
				if percent < 0.33 then
					r, g, b = unpack(db.colorHealthMin)
				elseif percent < 0.67 then
					r, g, b = unpack(db.colorHealthMid)
				else
					r, g, b = unpack(db.colorHealthMax)
				end
				color = ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
			end

			f.Health:SetFormattedText("%s%s|r", color, GetFormattedText(UnitIsPlayer(unit) and db.textFormatHealthPlayer or db.textFormatHealthMonster, cur, max))
			if db.alphaMode == "HEALTH" then
				if percent < 0.25 then
					f:SetAlpha(1)
				else
					f:SetAlpha(1.25 - percent)
				end
			end
		end
	end
end

function OnScreenHealth.UpdatePower()
	-- check options
	if not db.showPower then return end
	if not (db.playerEnable or db.targetEnable or db.petEnable) then return end
	local powerType
	for unit,_ in pairs(ns.units) do
		local f = unitFrames[unit]
		if ns.units[unit] and UnitExists(unit) and f and f:IsShown() then
			powerType = select(2, UnitPowerType(unit))
			if not realPowerTypes[powerType] then
				-- what is this, I don't even
				return
			end

			if UnitIsDeadOrGhost(unit) or not UnitIsConnected(unit) then
				f.Power:SetText()
			end
			local power,powermax = UnitPower(unit),UnitPowerMax(unit)

			local color = db.powerColors[powerType] or db.powerColors.MANA

			if powermax ~= 0 then
				f.Power:SetFormattedText("%s%s|r", color, GetFormattedText(UnitIsPlayer(unit) and db.textFormatPowerPlayer or db.textFormatPowerMonster, power, powermax))
			else
				f.Power:SetText()
			end
		end
	end
end

function OnScreenHealth.UpdateResource()
	-- check options
	if not db.showResource then return end
	if not (db.playerEnable or db.targetEnable or db.petEnable) then return end
	spec = spec or GetSpecialization()
	local unit = playerUnit
	for powerType, info in pairs(resourceTypes) do
		if not (powerType == "COMBO_POINTS" or powerType == "RUNES" or powerType == "DARK_FORCE") then  -- event based updates or exclusions
			if resourceTypes[powerType][playerClass] and ((spec and resourceTypes[powerType][playerClass] == spec) or (resourceTypes[powerType][playerClass] == 0)) and hasLevelReq and hasSpellReq then
				local power = UnitPower(unit, powerTypeID[powerType])
				if powerType == "ECLIPSE" then
					local direction = GetEclipseDirection()
					if direction then
						local color
						if power < 0 then
							color = db.resourceColors.ECLIPSE_LUNAR
						elseif power > 0 then
							color = db.resourceColors.ECLIPSE_SOLAR
						else
							color = "|cffffffff"
						end
						if direction == "moon" then
							unitFrames.player.Resource:SetFormattedText("< %s%d|r", color, abs(power))
						elseif direction == "sun" then
							unitFrames.player.Resource:SetFormattedText("%s%d|r >", color, abs(power))
						else
							unitFrames.player.Resource:SetFormattedText("%s%d|r", color, abs(power))
						end
					else
						unitFrames.player.Resource:SetText()
					end
				else
					if power and power ~= 0 then
						unitFrames.player.Resource:SetFormattedText("%s%d|r", db.resourceColors[powerType], abs(power))
					else
						unitFrames.player.Resource:SetText()
					end
				end
				return
			end
		end
	end
end

function OnScreenHealth:UpdateAll(type)
	if type and type == "health" then
		self.UpdateHealth()
	elseif type and type == "power" and db.showPower then
		self.UpdatePower()
	elseif type and type == "resource" and db.showResource then
		self.UpdateResource()
	else
		self:UpdatePlayer()
		self:PLAYER_TARGET_CHANGED()
		self:UNIT_PET("player")
	end
end

function OnScreenHealth:UpdatePlayer()
	if not db.playerEnable or not (db.playerShowOOC or (UnitAffectingCombat(playerUnit) or UnitAffectingCombat("pet")) or not (db.playerShowDead or not UnitIsDeadOrGhost(playerUnit))) then
		unitFrames.player:Hide()
		return
	end
	Debug(1, "Updating player")

	unitFrames.player:Show()

	if db.playerShowCombatStatus then
		self:UpdateCombatStatus(playerUnit)
	end
end

function OnScreenHealth:UpdateCombatStatus(unit)
	if not unit then unit = playerUnit end
	Debug(1, "Updating "..unit.." status")

	if db[unit.."ShowCombatStatus"] and (UnitAffectingCombat(unit) or UnitAffectingCombat("pet")) then
		unitFrames.player.Combat:SetText(db.colorCombatStatus.."+|r")
	else
		unitFrames.player.Combat:SetText()
	end
end

function OnScreenHealth.UpdateToT()
	if UnitExists("target") and UnitExists("targettarget") then
		if UnitIsUnit(playerUnit, "targettarget") then
			unitFrames.target.Target:SetText(db.colorTargetSelf..L["<<YOU>>"].."|r")
			return
		else
			unitFrames.target.Target:SetText(GetUnitColor("targettarget")..UnitName("targettarget").."|r")
			return
		end
	else
		unitFrames.target.Target:SetText()
	end
end

function OnScreenHealth:UpdateResourceEvents()
	Debug(1, "UpdateResourceEvents, showResource = %s", tostring(db.showResource))
	if db.showResource then
		if playerClass == "DEATHKNIGHT" then
			self:RegisterEvent("RUNE_POWER_UPDATE")
			self:RegisterEvent("RUNE_TYPE_UPDATE")
			self:RUNE_POWER_UPDATE()
		end
		if playerClass == "DRUID" or playerClass == "ROGUE" then
			self:RegisterEvent("UNIT_COMBO_POINTS")
			if playerClass == "DRUID" then
				self:RegisterEvent("UPDATE_SHAPESHIFT_FORM")
				self:UPDATE_SHAPESHIFT_FORM()
			end
			self:UNIT_COMBO_POINTS()
		end
	else
		self:UnregisterEvent("RUNE_POWER_UPDATE") -- death knight
		self:UnregisterEvent("RUNE_TYPE_UPDATE") -- death knight
		self:UnregisterEvent("UNIT_COMBO_POINTS") -- druid, rogue
		self:UnregisterEvent("UPDATE_SHAPESHIFT_FORM") -- druid
	end
end

function OnScreenHealth:UpdateCombatEvents()
	if db.playerShowOOC and db.targetShowOOC and db.alphaMode ~= "COMBAT" and not db.playerShowCombatStatus then
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	else
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
	end
end

function OnScreenHealth:UpdateDeathEvents()
	if db.playerShowDead then
		self:UnregisterEvent("PLAYER_ALIVE")
		self:UnregisterEvent("PLAYER_UNGHOST")
		self:UnregisterEvent("PLAYER_DEAD")
	else
		self:RegisterEvent("PLAYER_ALIVE")
		self:RegisterEvent("PLAYER_UNGHOST", "PLAYER_ALIVE")
		self:RegisterEvent("PLAYER_DEAD")
	end
end

-- --------------------------------------------------------------------
-- UI METHODS
-- --------------------------------------------------------------------
function OnScreenHealth:CreateFrames()
	Debug(1, "CreateFrames")
	local f, x, y

	if db.fontShadow then
		x, y = 1, -1
	else
		x, y = 0, 0
	end

	f = CreateFrame("Frame", "OnScreenHealthUpdateFrame", UIParent)
	self["updateframe"] = f

	f = CreateFrame("Frame", "OnScreenHealthPlayerFrame", UIParent)
	f:SetFrameStrata(db.frameStrata)
	f:SetFrameLevel(0)
	f:SetWidth(100)
	f:SetHeight(50)

	f.Health = f:CreateFontString(nil, "OVERLAY")
	f.Health:SetShadowOffset(x, y)

	f.Power = f:CreateFontString(nil, "OVERLAY")
	f.Power:SetShadowOffset(x, y)

	f.Combat = f:CreateFontString(nil, "OVERLAY")
	f.Combat:SetShadowOffset(x, y)

	f.Resource = f:CreateFontString(nil, "OVERLAY")
	f.Resource:SetShadowOffset(x, y)

	unitFrames["player"] = f
	unitFrames["vehicle"] = f

	f = CreateFrame("Frame", "OnScreenHealthPetFrame", UIParent)
	f:SetFrameStrata(db.frameStrata)
	f:SetFrameLevel(0)
	f:SetWidth(100)
	f:SetHeight(50)

	f.Health = f:CreateFontString(nil, "OVERLAY")
	f.Health:SetShadowOffset(x, y)

	f.Power = f:CreateFontString(nil, "OVERLAY")
	f.Power:SetShadowOffset(x, y)

	unitFrames["pet"] = f

	f = CreateFrame("Frame", "OnScreenHealthTargetFrame", UIParent)
	f:SetFrameStrata(db.frameStrata)
	f:SetFrameLevel(0)
	f:SetWidth(100)
	f:SetHeight(50)

	f.Health = f:CreateFontString(nil, "OVERLAY")
	f.Health:SetShadowOffset(x, y)

	f.Power = f:CreateFontString(nil, "OVERLAY")
	f.Power:SetShadowOffset(x, y)

	f.Target = f:CreateFontString(nil, "OVERLAY")
	f.Target:SetShadowOffset(x, y)

	f.ComboPoints = f:CreateFontString(nil, "OVERLAY")
	f.ComboPoints:SetShadowOffset(x, y)

	unitFrames["target"] = f

	local a
	if db.alphaMode == "COMBAT" then
		a = (UnitAffectingCombat(playerUnit) or UnitAffectingCombat("pet")) and db.alphaCombat or db.alphaOOC
	else
		a = db.alphaFixed
	end
	unitFrames["player"]:SetAlpha(a)
	unitFrames["pet"]:SetAlpha(a)
	unitFrames["target"]:SetAlpha(a)

	self:ApplyPositions()
	self:ApplyFonts()
	self:UpdatePlayer()
	self:PLAYER_TARGET_CHANGED()
	self:UNIT_PET("player")
end

function OnScreenHealth:ApplyPositions()
	Debug(1, "ApplyPositions")

	local p1, p2, p3, p4, p5, p6

	if db.textAlign == "CENTER" then
		p1, p2, p3, p4, p5, p6 = "TOP", "BOTTOM", "CENTER", "TOP", "BOTTOM", "CENTER"
	elseif db.textAlign == "OUTSIDE" then
		p1, p2, p3, p4, p5, p6 = "TOPLEFT", "BOTTOMLEFT", "LEFT", "TOPRIGHT", "BOTTOMRIGHT", "RIGHT"
	else
		p1, p2, p3, p4, p5, p6 = "TOPRIGHT", "BOTTOMRIGHT", "RIGHT", "TOPLEFT", "BOTTOMLEFT", "LEFT"
	end

	local f = unitFrames.player

	f:ClearAllPoints()
	f:SetPoint(p1, UIParent, "CENTER", -db.posX, db.posY)

	f.Health:ClearAllPoints()
	f.Health:SetPoint(p1, f, p1, 0, 0)
	f.Health:SetJustifyH(p3)

	f.Power:ClearAllPoints()
	f.Power:SetPoint(p1, f.Health, p2, 0, 0)
	f.Power:SetJustifyH(p3)

	f.Combat:ClearAllPoints()
	f.Combat:SetPoint(p2, f.Health, p1, 0, 0)
	f.Combat:SetJustifyH(p3)

	f.Resource:ClearAllPoints()
	f.Resource:SetPoint(p1, f.Power, p2, 0, 0)
	f.Resource:SetJustifyH(p3)

	f = unitFrames.target

	f:ClearAllPoints()
	f:SetPoint(p4, UIParent, "CENTER", db.posX, db.posY)

	f.Health:ClearAllPoints()
	f.Health:SetPoint(p4, f, p4, 0, 0)
	f.Health:SetJustifyH("CENTER")

	f.Power:ClearAllPoints()
	f.Power:SetPoint(p4, f.Health, p5, 0, 0)
	f.Power:SetJustifyH(p6)

	f.Target:ClearAllPoints()
	f.Target:SetPoint(p4, f.Power, p5, 0, 0)
	f.Target:SetJustifyH(p6)

	f.ComboPoints:ClearAllPoints()
	f.ComboPoints:SetPoint(p5, f.Health, p4, 0, 0)
	f.ComboPoints:SetJustifyH(p6)

	f = unitFrames.pet

	f:ClearAllPoints()
	f:SetPoint(p2, unitFrames.player, p1, 0, db.posPet)

	f.Health:ClearAllPoints()
	f.Health:SetPoint(p1, f, p1, 0, 0)
	f.Health:SetJustifyH(p3)

	f.Power:ClearAllPoints()
	f.Power:SetPoint(p1, f.Health, p2, 0, 0)
	f.Power:SetJustifyH(p3)
end

function OnScreenHealth:ApplyFonts()
	Debug(1, "ApplyFonts")

	local face = SharedMedia and SharedMedia:Fetch("font", db.fontFace) or "Fonts\\FRIZQT__.ttf"
	local s1, s2, outline = db.fontSizeHealth, db.fontSizePower, db.fontOutline

	local f = unitFrames.player
	f.Health:SetFont(face, s1, outline)
	f.Power:SetFont(face, s2, outline)
	f.Combat:SetFont(face, s2, outline)
	f.Resource:SetFont(face, s2, outline)

	f = unitFrames.pet
	f.Health:SetFont(face, s1 * db.fontSizePet, outline)
	f.Power:SetFont(face, s2 * db.fontSizePet, outline)

	f = unitFrames.target
	f.Health:SetFont(face, s1, outline)
	f.Power:SetFont(face, s2, outline)
	f.Target:SetFont(face, db.fontSizeTarget, outline)
 	f.ComboPoints:SetFont(face, db.fontSizeResource, outline)
end

function OnScreenHealth:Refresh()
	Debug(1, "Refresh")
	db = self.db.profile

	self:UpdateResourceEvents()
	self:UpdateCombatEvents()
	self:UpdateDeathEvents()

	self:ApplyFonts()
	self:ApplyPositions()

	self:UpdateAll()
end

-- --------------------------------------------------------------------
-- UPDATE ENGINE
-- --------------------------------------------------------------------
function OnScreenHealth.OnUpdate(frame, elapsed)
	frame.timer = frame.timer + elapsed
	if frame.timer >= db.playerFrequentUpdates then
		if db.targetShowToT then
			local newToT = UnitExists("targettarget") and UnitName("targettarget") or "NONE"
			if currentToT ~= newToT then
				currentToT = newToT
				OnScreenHealth.UpdateToT()
			end
		end
		OnScreenHealth.UpdateHealth()
		OnScreenHealth.UpdatePower()
		OnScreenHealth.UpdateResource()
		frame.timer = 0
	end
end

------------------------------------------------------------------------
-- OnScreenHealth.units = ns.units
-- OnScreenHealth.unitFrames = unitFrames
--[[
Notes:
AlternatePowerBar.lua
ShardBar.lua
MonkHarmonyBar.lua
PaladinPowerBar.lua
PriestBar.lua
RuneFrame.lua
EclipseBarFrame.lua
CompactUnitFrame.lua
UnitFrame.lua
Constants.lua
todo.txt
]]